# -*- coding: utf-8 -*-

# Copyright (c) 2009 Andrey Vlasovskikh
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re, errno
from subprocess import Popen, PIPE, CalledProcessError
from threading import Thread

READ_BUFSIZE = 4096

def pipe(pipeline='', *args, **opts):
    'str, ... > (Iterable(bytes) -> Iterable(bytes))'
    opts = opts.copy()
    opts.setdefault('stdout', PIPE)
    cmd = _format(pipeline, args)
    return lambda input: _run_pipeline(cmd, input, **opts)

def runpipe(pipeline='', *args, **opts):
    'str, ... -> Iterable(bytes)'
    return pipe(pipeline, *args, **opts)([])

def o(*fs):
    '*[y -> z, x -> y, ..., a -> b] -> (a -> z)'
    return lambda x: reduce(lambda x, f: f(x), reversed(fs), x)

def _shell_escape(str):
    'str -> str'
    return re.sub(r'''([ \t'"\$])''', r'\\\1', str)

def _format(pipeline, args):
    r'''str, [str] -> str

    >>> _format('ls -l {} | grep {} | wc', ['foo 1', 'bar$BAZ'])
    'ls -l foo\\ 1 | grep bar\\$BAZ | wc'
    '''
    # No escaping for '{}' at this moment
    assert pipeline.count('{}') == len(args)
    return reduce(
        lambda str, arg: str.replace('{}', arg, 1),
        map(_shell_escape, args),
        pipeline)

def _run_pipeline(cmd, input, **opts):
    def write(fd, xs):
        try:
            for x in xs:
                fd.write(x)
        finally:
            fd.close()
    opts.update(dict(shell=True, stdin=PIPE))
    p = Popen(cmd, **opts)
    try:
        writer = Thread(target=write, args=(p.stdin, iter(input)))
        writer.start()
        try:
            if p.stdout is not None:
                for x in iter(lambda: p.stdout.read(READ_BUFSIZE), ''):
                    yield x
        finally:
            writer.join()
    finally:
        ret = p.wait()
        if ret != 0:
            raise CalledProcessError(ret, cmd)




# -*- coding: utf-8 -*-

from __future__ import with_statement
import sys
from itertools import islice
from subprocess import CalledProcessError
from nose.tools import eq_, assert_raises
from minimock import Mock

from iterpipes import pipe, runpipe

STDOUT_FILENO = 1

join = lambda xs: ''.join(xs).strip()

def test_basic_single():
    eq_(join(runpipe('ls -d /')), '/')

def test_basic_many():
    eq_(join(runpipe('echo foo | wc -l')), '1')

def test_huge_input():
    wc = pipe('wc -l')
    eq_(join(wc('%d\n' % x for x in xrange(10000))), '10000')

def test_make_pipe():
    lines = join(runpipe(r'echo a b c b a | tr {} \\n | sort | uniq', ' '))
    eq_(lines, 'a\nb\nc')

def test_really_huge_input():
    cat = pipe('cat')
    sum = 0
    BUFSIZE = 10000
    TIMES = 10000
    with open('/dev/zero', 'rb') as fd:
        for x in cat(islice(iter(lambda: fd.read(BUFSIZE), ''), TIMES)):
            sum += len(x)
    eq_(sum, BUFSIZE * TIMES)

def test_dont_touch_stdout():
    eq_(join(runpipe('ls -d /', stdout=STDOUT_FILENO)), '')

def test_nonexistent_command():
    assert_raises(CalledProcessError,
        lambda: join(runpipe('echo foo | bar --baz')))

if __name__ == '__main__':
    test_basic_single()




# -*- coding: utf-8 -*-

from distutils.core import setup

setup(
    name='iterpipes',
    version='0.1',
    py_modules=['iterpipes'],
    author='Andrey Vlasovskikh',
    author_email='andrey.vlasovskikh@gmail.com',
    description='A library for running shell pipelines using shell-like syntax',
    license='MIT',
    url='http://bitbucket.org/vlasovskikh/iterpipes/')

